/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.docker.complex;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.File;
import java.util.UUID;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.openepics.names.docker.ITUtil;
import org.openepics.names.docker.ITUtilNameElement;
import org.openepics.names.docker.ITUtilStructureElement;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.NameElement;
import org.openepics.names.rest.beans.element.NameElementCommand;
import org.openepics.names.rest.beans.element.StructureElement;
import org.openepics.names.rest.beans.element.StructureElementCommand;
import org.openepics.names.util.EnumUtil.NameChoice;
import org.testcontainers.containers.DockerComposeContainer;
import org.testcontainers.containers.wait.strategy.Wait;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

/**
 * Integration tests for Naming and PostgreSQL that make use of existing dockerization
 * with docker-compose.yml / Dockerfile.
 *
 * <p>
 * Focus of this class is names endpoint.
 * In particular, purpose is to test arrays of names.
 * </p>
 *
 * @author Lars Johansson
 */
@Testcontainers
class NamesMultipleIT {

    // names for each of
    //     system group                       + description + comment
    //     system group + device type + index + description + comment
    //     system                             + description + comment
    //     system       + device type + index + description + comment
    //     subsystem                          + description + comment
    //     subsystem    + device type + index + description + comment

    @Container
    public static final DockerComposeContainer<?> ENVIRONMENT =
        new DockerComposeContainer<>(new File("docker-compose-integrationtest.yml"))
            .waitingFor(ITUtil.NAMING, Wait.forLogMessage(".*Started NamingApplication.*", 1));

    private static UUID systemGroupAcc     = null;
    private static UUID systemRFQ          = null;
    private static UUID subsystem010PRL    = null;

    private static UUID disciplineCryo     = null;
    private static UUID deviceGroupCryo    = null;
    private static UUID deviceType_Cryo_FS = null;

    @BeforeAll
    public static void initAll() {
        // init system group, system, subsystem, discipline, device group, device type

        StructureElementCommand structureElement  = null;
        StructureElement approvedStructureElement = null;

        structureElement = new StructureElementCommand(
                null, Type.SYSTEMGROUP, null,
                "Accelerator", "Acc",
                "The ESS Linear Accelerator", "approved by alfio");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        systemGroupAcc = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.SYSTEM, systemGroupAcc,
                "Radio Frequency Quadrupole", "RFQ",
                "empty", "empty");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        systemRFQ = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.SUBSYSTEM, systemRFQ,
                "01 Phase Reference Line", "010PRL",
                "empty", "Approved by Daniel Piso");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        subsystem010PRL = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "Cryogenics", "Cryo",
                "empty", "empty");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        disciplineCryo = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.DEVICEGROUP, disciplineCryo,
                "empty", null,
                "empty", "empty");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        deviceGroupCryo = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.DEVICETYPE, deviceGroupCryo,
                "Flow Switch", "FS",
                "empty", "empty");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        deviceType_Cryo_FS = approvedStructureElement.getUuid();
    }

    @Test
    void checkCreate() {
        // purpose
        //     test conditions for create name
        //         not create itself
        //
        // what - combination of
        //     read        exists name
        //     read        is legacy name
        //     read        is valid to create
        //     read        validate create
        //
        // note
        //     with and without index

        try {
            NameElementCommand[] nameElements = null;
            UUID uuid = null;
            String value = null;

            nameElements = new NameElementCommand[] {
                    new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "011", "checkCreate 1.3.011", "checkCreate 1.3.011"),
                    new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "012", "checkCreate 1.3.012", "checkCreate 1.3.012"),
                    new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "013", "checkCreate 1.3.013", "checkCreate 1.3.013"),
                    new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "011", "checkCreate 2.3.011", "checkCreate 2.3.011"),
                    new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "012", "checkCreate 2.3.012", "checkCreate 2.3.012"),
                    new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "013", "checkCreate 2.3.013", "checkCreate 2.3.013"),
                    new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "011", "checkCreate 3.3.011", "checkCreate 3.3.011"),
                    new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "012", "checkCreate 3.3.012", "checkCreate 3.3.012"),
                    new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "013", "checkCreate 3.3.013", "checkCreate 3.3.013")
            };

            ITUtilNameElement.assertExists("RFQ:Cryo-FS-011", Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ:Cryo-FS-011", Boolean.FALSE);
            ITUtilNameElement.assertIsValidToCreate("RFQ:Cryo-FS-011", Boolean.TRUE);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.DELETE, Boolean.FALSE);

            uuid = deviceType_Cryo_FS;
            nameElements[3].setUuid(uuid);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.FALSE);
            nameElements[3].setUuid(null);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);

            uuid = nameElements[3].getParentsystemstructure();
            nameElements[3].setParentsystemstructure(null);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.FALSE);
            nameElements[3].setParentsystemstructure(uuid);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);

            uuid = nameElements[3].getParentdevicestructure();
            nameElements[3].setParentdevicestructure(null);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.FALSE);
            nameElements[3].setParentdevicestructure(uuid);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);

            value = nameElements[3].getIndex();
            nameElements[3].setIndex(null);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.FALSE);
            nameElements[3].setIndex(value);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);

            value = nameElements[3].getDescription();
            nameElements[3].setDescription(null);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.FALSE);
            nameElements[3].setDescription(value);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);

            value = nameElements[3].getComment();
            nameElements[3].setComment(null);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.FALSE);
            nameElements[3].setComment(value);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void create() {
        // purpose
        //     test create name
        //
        // what - combination of
        //     create      create names
        //     read        exists name
        //     read        is legacy name
        //     read        is valid to create
        //     read        validate create
        //
        // note
        //     with and without index

        try {
            NameElementCommand[] nameElements = null;
            NameElement[] createdNameElements = null;
            int nbrNames = -1;

            nameElements = new NameElementCommand[] {
                    new NameElementCommand(null, systemGroupAcc,  null,               null,  "create 1.0.0",   "create 1.0.0"),
                    new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "021", "create 1.3.021", "create 1.3.021"),
                    new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "022", "create 1.3.022", "create 1.3.022"),
                    new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "023", "create 1.3.023", "create 1.3.023"),
                    new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "021", "create 2.3.021", "create 2.3.021"),
                    new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "022", "create 2.3.022", "create 2.3.022"),
                    new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "023", "create 2.3.023", "create 2.3.023"),
                    new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "021", "create 3.3.021", "create 3.3.021"),
                    new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "022", "create 3.3.022", "create 3.3.022"),
                    new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "023", "create 3.3.023", "create 3.3.023")
            };

            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.DELETE, Boolean.FALSE);

            nameElements = new NameElementCommand[] {
                    nameElements[1],
                    nameElements[2],
                    nameElements[3],
                    nameElements[4],
                    nameElements[5],
                    nameElements[6],
                    nameElements[7],
                    nameElements[8],
                    nameElements[9]
            };

            ITUtilNameElement.assertExists("Acc:Cryo-FS-023", Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("Acc:Cryo-FS-023", Boolean.FALSE);
            ITUtilNameElement.assertIsValidToCreate("Acc:Cryo-FS-023", Boolean.TRUE);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElements, NameChoice.DELETE, Boolean.FALSE);

            nbrNames = ITUtilNameElement.assertRead("?deleted=false").getListSize();

            createdNameElements = ITUtilNameElement.assertCreate(nameElements);
            assertNotNull(createdNameElements);
            assertEquals(9, createdNameElements.length);

            assertEquals(nbrNames + 9, ITUtilNameElement.assertRead("?deleted=false").getListSize());

            ITUtilNameElement.assertExists("Acc:Cryo-FS-023", Boolean.TRUE);
            ITUtilNameElement.assertIsLegacy("Acc:Cryo-FS-023", Boolean.FALSE);
            ITUtilNameElement.assertIsValidToCreate("Acc:Cryo-FS-023", Boolean.FALSE);
            ITUtilNameElement.assertValidate(createdNameElements, NameChoice.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);
            ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkUpdate() {
        // purpose
        //     test conditions for update name
        //         not update
        //
        // what - combination of
        //     create      create names
        //     read        exists name
        //     read        is legacy name
        //     read        is valid to create
        //     read        validate create
        //     read        validate update
        //
        // note
        //     create in order to update
        //     validate update

        NameElementCommand[] nameElements = null;
        NameElement[] createdNameElements = null;
        int nbrNames = -1;
        UUID uuid = null;
        String value = null;

        nameElements = new NameElementCommand[] {
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "031", "checkUpdate 1.3.031", "checkUpdate 1.3.031"),
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "032", "checkUpdate 1.3.032", "checkUpdate 1.3.032"),
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "033", "checkUpdate 1.3.033", "checkUpdate 1.3.033"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "031", "checkUpdate 2.3.031", "checkUpdate 2.3.031"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "032", "checkUpdate 2.3.032", "checkUpdate 2.3.032"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "033", "checkUpdate 2.3.033", "checkUpdate 2.3.033"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "031", "checkUpdate 3.3.031", "checkUpdate 3.3.031"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "032", "checkUpdate 3.3.032", "checkUpdate 3.3.032"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "033", "checkUpdate 3.3.033", "checkUpdate 3.3.033")
        };

        ITUtilNameElement.assertExists("RFQ:Cryo-FS-031", Boolean.FALSE);
        ITUtilNameElement.assertIsLegacy("RFQ:Cryo-FS-031", Boolean.FALSE);
        ITUtilNameElement.assertIsValidToCreate("RFQ:Cryo-FS-031", Boolean.TRUE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.UPDATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.DELETE, Boolean.FALSE);

        nbrNames = ITUtilNameElement.assertRead("?deleted=false").getListSize();

        createdNameElements = ITUtilNameElement.assertCreate(nameElements);
        assertNotNull(createdNameElements);
        assertEquals(9, createdNameElements.length);

        assertEquals(nbrNames + 9, ITUtilNameElement.assertRead("?deleted=false").getListSize());

        ITUtilNameElement.assertExists("RFQ:Cryo-FS-031", Boolean.TRUE);
        ITUtilNameElement.assertIsLegacy("RFQ:Cryo-FS-031", Boolean.FALSE);
        ITUtilNameElement.assertIsValidToCreate("RFQ:Cryo-FS-031", Boolean.FALSE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.CREATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);

        uuid = createdNameElements[3].getUuid();
        createdNameElements[3].setUuid(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.FALSE);
        createdNameElements[3].setUuid(uuid);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);

        uuid = createdNameElements[3].getParentsystemstructure();
        createdNameElements[3].setParentsystemstructure(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.FALSE);
        createdNameElements[3].setParentsystemstructure(uuid);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);

        uuid = createdNameElements[3].getParentdevicestructure();
        createdNameElements[3].setParentdevicestructure(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.FALSE);
        createdNameElements[3].setParentdevicestructure(uuid);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);

        value = createdNameElements[3].getIndex();
        createdNameElements[3].setIndex(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.FALSE);
        createdNameElements[3].setIndex(value);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);

        value = createdNameElements[3].getDescription();
        createdNameElements[3].setDescription(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.FALSE);
        createdNameElements[3].setDescription(value);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);

        value = createdNameElements[3].getComment();
        createdNameElements[3].setComment(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.FALSE);
        createdNameElements[3].setComment(value);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);
    }

    @Test
    void update() {
        // purpose
        //     test update name
        //
        // what - combination of
        //     create      create names
        //     read        exists name
        //     read        is legacy name
        //     read        is valid to create
        //     read        validate create
        //     read        validate update
        //     update      update names
        //
        // note
        //     create in order to update

        NameElementCommand[] nameElements = null;
        NameElement[] createdNameElements = null;
        NameElement[] updatedNameElements = null;
        int nbrNames = -1;

        nameElements = new NameElementCommand[] {
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "041", "update 1.3.041", "update 1.3.041"),
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "042", "update 1.3.042", "update 1.3.042"),
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "043", "update 1.3.043", "update 1.3.043"),
                new NameElementCommand(null, systemRFQ,       null,               null,  "update 2.0.0",   "update 2.0.0"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "041", "update 2.3.041", "update 2.3.041"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "042", "update 2.3.042", "update 2.3.042"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "043", "update 2.3.043", "update 2.3.043"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "041", "update 3.3.041", "update 3.3.041"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "042", "update 3.3.042", "update 3.3.042"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "043", "update 3.3.043", "update 3.3.043")
        };

        ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.UPDATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.DELETE, Boolean.FALSE);

        nameElements = new NameElementCommand[] {
                nameElements[0],
                nameElements[1],
                nameElements[2],
                nameElements[4],
                nameElements[5],
                nameElements[6],
                nameElements[7],
                nameElements[8],
                nameElements[9]
        };

        ITUtilNameElement.assertExists("RFQ", Boolean.TRUE);
        ITUtilNameElement.assertIsLegacy("RFQ", Boolean.FALSE);
        ITUtilNameElement.assertIsValidToCreate("RFQ", Boolean.FALSE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.UPDATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.DELETE, Boolean.FALSE);

        nbrNames = ITUtilNameElement.assertRead("?deleted=false").getListSize();

        createdNameElements = ITUtilNameElement.assertCreate(nameElements);
        assertNotNull(createdNameElements);
        assertEquals(9, createdNameElements.length);

        assertEquals(nbrNames + 9, ITUtilNameElement.assertRead("?deleted=false").getListSize());

        ITUtilNameElement.assertExists("RFQ", Boolean.TRUE);
        ITUtilNameElement.assertIsLegacy("RFQ", Boolean.FALSE);
        ITUtilNameElement.assertIsValidToCreate("RFQ", Boolean.FALSE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.CREATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);

        createdNameElements[0].setDescription("0");
        createdNameElements[1].setDescription("1");
        createdNameElements[2].setDescription("2");
        createdNameElements[3].setDescription("3");
        createdNameElements[4].setDescription("4");
        createdNameElements[5].setDescription("5");
        createdNameElements[6].setDescription("6");
        createdNameElements[7].setDescription("7");
        createdNameElements[8].setDescription("8");

        ITUtilNameElement.assertExists("RFQ", Boolean.TRUE);
        ITUtilNameElement.assertIsLegacy("RFQ", Boolean.FALSE);
        ITUtilNameElement.assertIsValidToCreate("RFQ", Boolean.FALSE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.CREATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);

        updatedNameElements = ITUtilNameElement.assertUpdate(createdNameElements);
        assertNotNull(updatedNameElements);
        assertEquals(9, updatedNameElements.length);

        assertEquals(nbrNames + 9, ITUtilNameElement.assertRead("?deleted=false").getListSize());

        ITUtilNameElement.assertExists("RFQ", Boolean.TRUE);
        ITUtilNameElement.assertIsLegacy("RFQ", Boolean.FALSE);
        ITUtilNameElement.assertIsValidToCreate("RFQ", Boolean.FALSE);
        ITUtilNameElement.assertValidate(updatedNameElements, NameChoice.CREATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(updatedNameElements, NameChoice.UPDATE, Boolean.TRUE);
        ITUtilNameElement.assertValidate(updatedNameElements, NameChoice.DELETE, Boolean.TRUE);
    }

    @Test
    void checkDelete() {
        // purpose
        //     test conditions for delete name
        //         not delete
        //
        // what - combination of
        //     create      create names
        //     read        exists name
        //     read        is legacy name
        //     read        is valid to create
        //     read        validate create
        //     read        validate delete
        //
        // note
        //     create in order to delete
        //     validate delete
        //         uuid
        //         comment

        NameElementCommand[] nameElements = null;
        NameElement[] createdNameElements = null;
        int nbrNames = -1;
        UUID uuid = null;
        String value = null;

        nameElements = new NameElementCommand[] {
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "051", "checkDelete 1.3.051", "checkDelete 1.3.051"),
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "052", "checkDelete 1.3.052", "checkDelete 1.3.052"),
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "053", "checkDelete 1.3.053", "checkDelete 1.3.053"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "051", "checkDelete 2.3.051", "checkDelete 2.3.051"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "052", "checkDelete 2.3.052", "checkDelete 2.3.052"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "053", "checkDelete 2.3.053", "checkDelete 2.3.053"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "051", "checkDelete 3.3.051", "checkDelete 3.3.051"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "052", "checkDelete 3.3.052", "checkDelete 3.3.052"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "053", "checkDelete 3.3.053", "checkDelete 3.3.053")
        };

        ITUtilNameElement.assertExists("RFQ:Cryo-FS-051", Boolean.FALSE);
        ITUtilNameElement.assertIsLegacy("RFQ:Cryo-FS-051", Boolean.FALSE);
        ITUtilNameElement.assertIsValidToCreate("RFQ:Cryo-FS-051", Boolean.TRUE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.UPDATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.DELETE, Boolean.FALSE);

        nbrNames = ITUtilNameElement.assertRead("?deleted=false").getListSize();

        createdNameElements = ITUtilNameElement.assertCreate(nameElements);
        assertNotNull(createdNameElements);
        assertEquals(9, createdNameElements.length);

        assertEquals(nbrNames + 9, ITUtilNameElement.assertRead("?deleted=false").getListSize());

        ITUtilNameElement.assertExists("RFQ:Cryo-FS-051", Boolean.TRUE);
        ITUtilNameElement.assertIsLegacy("RFQ:Cryo-FS-051", Boolean.FALSE);
        ITUtilNameElement.assertIsValidToCreate("RFQ:Cryo-FS-051", Boolean.FALSE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.CREATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);

        uuid = createdNameElements[3].getUuid();
        createdNameElements[3].setUuid(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.FALSE);
        createdNameElements[3].setUuid(uuid);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);

        uuid = createdNameElements[3].getParentsystemstructure();
        createdNameElements[3].setParentsystemstructure(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);
        createdNameElements[3].setParentsystemstructure(uuid);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);

        uuid = createdNameElements[3].getParentdevicestructure();
        createdNameElements[3].setParentdevicestructure(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);
        createdNameElements[3].setParentdevicestructure(uuid);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);

        value = createdNameElements[3].getIndex();
        createdNameElements[3].setIndex(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);
        createdNameElements[3].setIndex(value);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);

        value = createdNameElements[3].getDescription();
        createdNameElements[3].setDescription(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);
        createdNameElements[3].setDescription(value);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);

        value = createdNameElements[3].getComment();
        createdNameElements[3].setComment(null);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.FALSE);
        createdNameElements[3].setComment(value);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);
    }

    @Test
    void delete() {
        // purpose
        //     test delete name
        //
        // what - combination of
        //     create      create names
        //     read        exists name
        //     read        is legacy name
        //     read        is valid to create
        //     read        validate create
        //     read        validate update
        //     delete      delete names
        //
        // note
        //     create in order to delete
        //
        // names for each of
        //     system group                       + description + comment
        //     system group + device type + index + description + comment
        //     system                             + description + comment
        //     system       + device type + index + description + comment
        //     subsystem                          + description + comment
        //     subsystem    + device type + index + description + comment

        NameElementCommand[] nameElements = null;
        NameElement[] createdNameElements = null;
        NameElement[] deletedNameElements = null;
        int nbrNames = -1;

        nameElements = new NameElementCommand[] {
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "061", "delete 1.3.061", "delete 1.3.061"),
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "062", "delete 1.3.062", "delete 1.3.062"),
                new NameElementCommand(null, systemGroupAcc,  deviceType_Cryo_FS, "063", "delete 1.3.063", "delete 1.3.063"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "061", "delete 2.3.061", "delete 2.3.061"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "062", "delete 2.3.062", "delete 2.3.062"),
                new NameElementCommand(null, systemRFQ,       deviceType_Cryo_FS, "063", "delete 2.3.063", "delete 2.3.063"),
                new NameElementCommand(null, subsystem010PRL, null,               null,  "delete 3.0.0",   "delete 3.0.0"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "061", "delete 3.3.061", "delete 3.3.061"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "062", "delete 3.3.062", "delete 3.3.062"),
                new NameElementCommand(null, subsystem010PRL, deviceType_Cryo_FS, "063", "delete 3.3.063", "delete 3.3.063")
        };

        ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.UPDATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.DELETE, Boolean.FALSE);

        nameElements = new NameElementCommand[] {
                nameElements[0],
                nameElements[1],
                nameElements[2],
                nameElements[3],
                nameElements[4],
                nameElements[5],
                nameElements[7],
                nameElements[8],
                nameElements[9]
        };

        ITUtilNameElement.assertExists("RFQ:Cryo-FS-061", Boolean.FALSE);
        ITUtilNameElement.assertIsLegacy("RFQ:Cryo-FS-061", Boolean.FALSE);
        ITUtilNameElement.assertIsValidToCreate("RFQ:Cryo-FS-061", Boolean.TRUE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.UPDATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.DELETE, Boolean.FALSE);

        nbrNames = ITUtilNameElement.assertRead("?deleted=false").getListSize();

        createdNameElements = ITUtilNameElement.assertCreate(nameElements);
        assertNotNull(createdNameElements);
        assertEquals(9, createdNameElements.length);

        assertEquals(nbrNames + 9, ITUtilNameElement.assertRead("?deleted=false").getListSize());

        ITUtilNameElement.assertExists("RFQ:Cryo-FS-061", Boolean.TRUE);
        ITUtilNameElement.assertIsLegacy("RFQ:Cryo-FS-061", Boolean.FALSE);
        ITUtilNameElement.assertIsValidToCreate("RFQ:Cryo-FS-061", Boolean.FALSE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.CREATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.UPDATE, Boolean.TRUE);
        ITUtilNameElement.assertValidate(createdNameElements, NameChoice.DELETE, Boolean.TRUE);

        deletedNameElements = ITUtilNameElement.assertDelete(createdNameElements);
        assertNotNull(deletedNameElements);
        assertEquals(9, deletedNameElements.length);

        assertEquals(nbrNames, ITUtilNameElement.assertRead("?deleted=false").getListSize());

        ITUtilNameElement.assertExists("RFQ:Cryo-FS-061", Boolean.FALSE);
        ITUtilNameElement.assertIsLegacy("RFQ:Cryo-FS-061", Boolean.FALSE);
        ITUtilNameElement.assertIsValidToCreate("RFQ:Cryo-FS-061", Boolean.TRUE);
        ITUtilNameElement.assertValidate(nameElements, NameChoice.CREATE, Boolean.TRUE);
        ITUtilNameElement.assertValidate(deletedNameElements, NameChoice.CREATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(deletedNameElements, NameChoice.UPDATE, Boolean.FALSE);
        ITUtilNameElement.assertValidate(deletedNameElements, NameChoice.DELETE, Boolean.FALSE);
    }

}
